// **************************************************************************
// * This file is part of the FreeFileSync project. It is distributed under *
// * GNU General Public License: http://www.gnu.org/licenses/gpl.html       *
// * Copyright (C) Zenju (zenju AT gmx DOT de) - All Rights Reserved        *
// **************************************************************************

#include <iostream>
#include <fstream>
#include <string>
#include <iterator>
#include <zen/string_tools.h>


using namespace zen;

//reads master translation file and adapts target language file accordingly, leaving not translated texts empty (which leads to an NSIS syntax error
//thous ensures that all texts are actually translated


//sample translation line:
//      LangString Lang_Help ${LANG_ENGLISH} "Help" ;comment
//      \____ Prefix ______/  \_Identifier_/  \_ Localized text _/

struct TranslationLine
{
    std::string prefix;
    std::string localization;
};


class NsisError
{
public:
    explicit NsisError(const std::string& msg) : msg_(msg) {}
    std::string toString() const { return msg_; }

private:
    std::string msg_;
};


void loadFile(const std::string&        filename,
              std::vector<std::string>& header,
              std::string&              identifier,
              std::vector<TranslationLine>& textBuffer) //throw NsisError
{
    header.clear();
    identifier.clear();
    textBuffer.clear();

    std::ifstream inputFile(filename.c_str());
    if (!inputFile)
        throw NsisError("Error opening file for reading: " + filename);

    std::vector<char> buffer(10000);

    bool processingHeader = true;

    while (inputFile.getline(&buffer[0], buffer.size()))
    {
        //remove UTF-8 Byte Ordering Mark
        if (    buffer[0] == static_cast<char>(0xEF) &&
                buffer[1] == static_cast<char>(0xBB) &&
                buffer[2] == static_cast<char>(0xBF))
        {
            buffer.erase(buffer.begin(), buffer.begin() + 3);
        }

        std::string newLine = &buffer[0];
        trim(newLine, true, false);

        if (startsWith(newLine, "LangString"))
        {
            processingHeader = false; //header ends with first occurence of "LangString"

            const size_t loc = newLine.find("${");
            if (loc != std::string::npos)
            {
                const size_t loc2 = newLine.find("}", loc + 2);
                if (loc2 != std::string::npos)
                {
                    const std::string newIdentifier = std::string(newLine.c_str() + loc, loc2 - loc + 1);

                    //set or test language id
                    if (identifier.empty())
                        identifier = newIdentifier;
                    else
                    {
                        if (identifier != newIdentifier)
                            throw NsisError("Invalid file: different language identifiers: " + filename);
                    }

                    TranslationLine newEntry;
                    newEntry.prefix       = std::string(newLine.c_str(), loc);
                    newEntry.localization = std::string(newLine.c_str() + loc2 + 1);
                    textBuffer.push_back(newEntry);
                    continue;
                }
            }

            throw NsisError("Invalid file: not all rows contain valid language identifier: " + filename);
        }
        else if (processingHeader && !startsWith(newLine, ";")) //don't preserve comments in header
            header.push_back(newLine);
    }

    //consistency checks
    if (header.empty())
        throw NsisError("Invalid file: header is empty: " + filename);

    if (identifier.empty())
        throw NsisError("Invalid file: language identifier not found: " + filename);

    if (textBuffer.empty())
        throw NsisError("Invalid file: no translation relevant rows found: " + filename);
}


int main(int args, char** argv)
{
    if (args != 3)
    {
        std::cout << "\nInvalid input parameters: Specify master language file, e.g. \"english.nsh\", as (%1) and changing language file (%2)\n" << std::endl;
        system("pause");
        return -1;
    }

    const std::string masterFileName   = argv[1];
    const std::string changingFileName = argv[2];

    if (afterLast(masterFileName, '\\') == afterLast(changingFileName, '\\')) //input == changing file
        return 0;

    //master file
    std::vector<std::string> masterHeader;
    std::string masterIdentifier;
    std::vector<TranslationLine> masterTextBuffer;

    //changing file
    std::vector<std::string> changingHeader;
    std::string changingIdentifier;
    std::vector<TranslationLine> changingTextBuffer;

    try
    {
        loadFile(masterFileName,
                 masterHeader,
                 masterIdentifier,
                 masterTextBuffer); //throw NsisError

        loadFile(changingFileName,
                 changingHeader,
                 changingIdentifier,
                 changingTextBuffer); //throw NsisError

        if (masterIdentifier == changingIdentifier)
            throw NsisError("Both files have the same language identifier!");
    }
    catch (const NsisError& e)
    {
        std::cout << e.toString() << std::endl;
        system("pause");
        return -1;
    }

    //update changing file
    std::ofstream output(changingFileName.c_str());

    //write header
    std::copy(changingHeader.begin(), changingHeader.end(), std::ostream_iterator<std::string>(output, "\n"));

    bool notTranslatedTextsExist = false;
    for (std::vector<TranslationLine>::const_iterator iterMaster = masterTextBuffer.begin(); iterMaster != masterTextBuffer.end(); ++iterMaster)
    {
        std::string localization;

        auto iter = std::find_if(changingTextBuffer.begin(), changingTextBuffer.end(), [&](const TranslationLine& tl) { return tl.prefix == iterMaster->prefix; });
        if (iter != changingTextBuffer.end())
        {
            std::string tmp = iter->localization;
            trim(tmp);
            if (!tmp.empty())
                localization = iter->localization;
        }

        if (localization.empty())
            notTranslatedTextsExist = true; //leave newEntry.localized empty!

        //write body
        output << ";source text -> " << iterMaster->localization << "\n"; //show original text to help translators
        output << iterMaster->prefix << changingIdentifier << localization << "\n\n";
    }

    if (notTranslatedTextsExist)
        std::cout << "\nTexts to be translated are existing!!!" << std::endl << std::endl;
    else
        std::cout << "All OK!" << std::endl << std::endl;

    return 0;
}
